let customBadges = {}

// the slime2:ready event is fired once
// indicates that the slime2 global variable is ready to use
addEventListener('slime2:ready', () => {
  slime2.widget.loadPlatform('twitch')

  slime2.onEvent(event => {
    switch (event.type) {
      case 'message':
        return onMessage(event.message)
    }
  })
})

// this event is fired every time a widget setting is changed
// use this to update the widget live, rather than the user having to refresh
addEventListener('slime2:widget-data-update', () => {
  // slime2.widget.getData() will always fetch the latest data
  const data = slime2.widget.getData()

  // pull settings
  const {
    stacks,
    general,
    botFilters,
    userFilters,
    messageFilters,
    fontStyle,
    fullStyle,
    compactStyle,
  } = data
  
  // pull general settings
  const {
    position,
    timing,
    animations,
    badges,
    emotes,
    pronouns,    
  } = general
  
  // pull full message settings
  const {
    fullUser,
    fullMessage,    
  } = fullStyle
  
  const eventList = $('#slime2-event-list').removeClass()

  // stack settings
  eventList.addClass(`${stacks}`)
  
  // position settings
  eventList.addClass(`${position.direction}-${position.corner}`)

  // timing settings
  slime2.setEventDelay(timing.delay * 1000)
  slime2.setMaxEvents(timing.max)
  slime2.setEventExpiration(
    timing.expiration ? timing.expiration * 1000 : undefined,
    animations.exit !== 'none'
      ? {
          animationTime: 500,
          animationClass: `animation-exit-${animations.exit}`,
        }
      : undefined,
  )

  // badge settings
  if (!badges.display) eventList.addClass('hide-badges')
  eventList.css({
    '--badgeSize': `${badges.size}px`,
  })
  
  customBadges = {
    broadcaster: badges.hostBadge,
    moderator: badges.modBadge,
    vip: badges.vipBadge,
    founder: badges.founderBadge,
    'artist-badge': badges.artistBadge,
  }
  
  // emote settings
  if (emotes.dynamic)
    eventList.addClass('dynamic-emote-sizes').css({
      '--emoteMedium': `${emotes.medium}px`,
      '--emoteLarge': `${emotes.large}px`,
    })

  // pronouns settings
  eventList.addClass(`pronouns-${pronouns.display}`)  

  // style settings
  eventList
    .addClass([
      `text-edge-${fontStyle.outline}-${fontStyle.outlineSection}`,
      `${stacks}-user-color-${fullUser.userBackgroundColor}`,
      `${stacks}-user-color-${compactStyle.userColor}`,
    ])
    .css({
      // font settings
      '--fontName': `${fontStyle.font}`,
      '--fontSize': `${fontStyle.size}px`,
      '--fontWeight': fontStyle.weight,
      '--outlineColor': fontStyle.outlineColor,
      
      // full message settings
      '--fullUserBG': fullUser.customUserColor,
      '--fullUserText': fullUser.customUserTextColor,
      '--userAngle': `${fullUser.userAngle}deg`,
      '--userBorderRadius': `${fullUser.userBorderRadius}px`,
      '--userBorderWidth': `${fullUser.userBorderWidth}px`,
      '--userBorderColor': fullUser.userBorderColor,
      '--userVertical': `${fullUser.userVertical}em`,
      '--userHorizontal': `${fullUser.userHorizontal}em`,

      '--fullMessageBG': fullMessage.customMessageColor,
      '--fullMessageText': fullMessage.customMessageTextColor,
      '--messageBorderRadius': `${fullMessage.messageBorderRadius}px`,
      '--messageBorderWidth': `${fullMessage.messageBorderWidth}px`,
      '--messageBorderColor': fullMessage.messageBorderColor,
      '--messageHorizontal': `${fullMessage.messageHorizontal}em`,
      
      // compact message settings
      '--compactUserText': compactStyle.customUserColor,
      '--compactMessageText': compactStyle.customMessageColor,
      '--compactBG': compactStyle.customBackgroundColor,
      '--compactBorderRadius': `${compactStyle.compactBorderRadius}px`,
      '--compactBorderWidth': `${compactStyle.compactBorderWidth}px`,
      '--compactBorderColor': compactStyle.compactBorderColor,      

      // 20% opacity of edgeColor
      '--outlineColorLowOpacity': slime2.color
        .mix('transparent', fontStyle.outlineColor, 0.2, {
          space: 'srgb',
        })
        .toString({ format: 'rgba' }),
    })
    
    // adding compact padding if background and border are present
    const compactBG = compactStyle.customBackgroundColor;
    const compactBW = compactStyle.compactBorderWidth;
    const compactBC = compactStyle.compactBorderColor;
    if (hasOpacity(compactBG) || (compactBW > 0 && hasOpacity(compactBC))){
      eventList.addClass(`compact-background-on`)
    }
})

/****************
 * Chat Handler *
 ****************/
function onMessage(message) {

  if (!evaluateFilters(message)) return
  const animations = slime2.widget.getData([{id: 'general'}], 'animations')

  // extract the necessary data from the message
  const { parts, user } = message

  // clone the main message template to insert the message data into
  // the templates are all defined in the HTML file
  const messageClone = $(slime2.cloneTemplate('message-template'))

  // create the elements for the user and message content within the clone
  // this uses the element builder functions defined below
  messageClone.find('.user').append(buildUser(user))
  messageClone.find('.content').append(buildContent(parts))

  // add animation
  messageClone.find('.message').addClass(`animation-enter-${animations.enter}`)

  // add user's name color
  messageClone.find('.message').css({
    '--userColor': user.color,
    '--userColorLight': slime2.color.light(user.color),
    '--userColorDark': slime2.color.dark(user.color),
  })

  // set emote size
  const emoteSize = calculateEmoteSize(parts)
  messageClone.find('.content').addClass(`emote-${emoteSize}`)

  // return the message fragment which renders the fragment
  return {
    fragment: messageClone,
  }
}

/********************
 * Element Builders *
 ********************/

// insert in the user data
function buildUser(user) {
  const { displayName, userName, badges = [], pronouns } = user

  const userClone = $(slime2.cloneTemplate('user-template'))
  userClone.find('.badges').append(buildBadges(badges))

  let name = displayName

  // handle localized display names
  // https://blog.twitch.tv/en/2016/08/22/localized-display-names-e00ee8d3250a/
  if (userName && displayName.toLowerCase() !== userName.toLowerCase()) {
    name = `${displayName} (${userName})`
  }

  // insert in the user's pronouns and display name
  userClone.find('.pronouns').text(pronouns)
  userClone.find('.name').text(name)

  return userClone
}

// go through every message part, inserting in the emote, cheer, or text data
function buildContent(parts) {
  return parts.map(part => {
    switch (part.type) {
      case 'emote':
        return buildEmote(part)

      case 'cheer':
        return buildCheer(part)

      case 'text':
      default:
        return buildText(part)
    }
  })
}

// insert in the badge images
function buildBadges(badges) {
  return badges.map(badge => {
    const { id, image } = badge
    const badgeClone = $(slime2.cloneTemplate('badge-template'))
    
    if (customBadges && customBadges[id]) {
      badgeClone.find('.badge').attr('src', customBadges[id])
    }
    else {
      badgeClone.find('.badge').attr('src', image)
    }

    return badgeClone
  })
}

// insert in the emote image
function buildEmote(part) {
  const emotes = slime2.widget.getData([{id: 'general'}], 'emotes')
  const { images } = part.emote

  const emoteClone = $(slime2.cloneTemplate('content-emote-template'))
  emoteClone
    .find('.emote')
    .attr('src', emotes.static ? images.static.x4 : images.default.x4)

  return emoteClone
}

// insert in the cheermote image and cheer amount
function buildCheer(part) {
  const emotes = slime2.widget.getData([{id: 'general'}], 'emotes')
  const { amount, color, images } = part.cheer

  const cheerClone = $(slime2.cloneTemplate('content-cheer-template'))
  cheerClone
    .find('.emote')
    .attr('src', emotes.static ? images.static.x4 : images.default.x4)
  cheerClone
    .find('.cheer-amount')
    .text(amount)
    .css('color', color)

  return cheerClone
}

// insert in the text
function buildText(part) {
  const { text } = part

  const textClone = $(slime2.cloneTemplate('content-text-template'))
  textClone.find('.text').text(text)

  return textClone
}

/***********
 * Filters *
 ***********/

function evaluateFilters(message) {
  return (
    evaluateBotFilters(message) &&
    evaluateUserFilters(message) &&
    evaluateMessageFilters(message)
  )
}

function evaluateBotFilters(message) {
  const { botFilters } = slime2.widget.getData().filters
  const { text, user } = message

  for (const prefix of botFilters.prefixes) {
    if (text.startsWith(prefix)) return false
  }

  for (const name of botFilters.names) {
    if (user.userName.toLowerCase() === name.toLowerCase()) return false
  }

  return true
}

function evaluateUserFilters(message) {
  const { userFilters } = slime2.widget.getData().filters
  const { user } = message

  for (const name of userFilters.users) {
    if (user.userName.toLowerCase() === name.toLowerCase()) return true
  }

  const typeMap = new Map(userFilters.types.map(type => [type, true]))

  if (typeMap.get('all')) return true

  const { roles } = user
  if (typeMap.get('subs') && roles.subscriber) return true
  if (typeMap.get('mods') && roles.moderator) return true
  if (typeMap.get('vips') && roles.vip) return true
  if (typeMap.get('artists') && roles.artist) return true

  if (typeMap.get('followers')) {
    const { followDate } = user
    if (!followDate) return false

    const minFollowTime = (userFilters.followHours || 0) * 60 * 60 * 1000
    const userFollowTime = Date.now() - followDate.getTime()

    return userFollowTime > minFollowTime
  }

  return false
}

function evaluateMessageFilters(message) {
  const { messageFilters } = slime2.widget.getData().filters

  for (const word of messageFilters.words) {
    if (message.text.includes(word)) return false
  }

  const typeMap = new Map(messageFilters.types.map(type => [type, true]))
  if (!typeMap.size) return true

  if (typeMap.get('text') && message.parts.some(part => part.type === 'text')) {
    return false
  }

  if (
    typeMap.get('emote') &&
    message.parts.some(part => part.type !== 'text')
  ) {
    return false
  }

  if (typeMap.get('first') && message.first) return false

  const { type } = message
  if (typeMap.get('action') && type === 'action') return false
  if (typeMap.get('cheer') && type === 'cheer') return false
  if (typeMap.get('reply') && type === 'reply') return false
  if (typeMap.get('highlight') && type === 'highlight') return false
  if (typeMap.get('redeem') && type === 'redeem') return false
  if (typeMap.get('resub') && type === 'resub') return false
  if (typeMap.get('announcement') && type === 'announcement') return false

  return true
}

/********************
 * Helper Functions *
 ********************/

// emote size is 1 if the message contains anything other than emotes
// 2 if the message contains multiple emotes and nothing else
// 4 if the entire messages is just a single emote
function calculateEmoteSize(parts) {
  let emoteCount = 0
  let otherCount = 0

  parts.forEach(part => {
    if (part.type === 'emote') {
      emoteCount++
    } else if (part.text.trim() !== '') {
      otherCount++
    }
  })

  if (otherCount > 0) {
    return 1
  } else if (emoteCount > 1) {
    return 2
  }

  return 4
}


// transparency check
function hasOpacity(colorString) {
  const color = new slime2.color(colorString)
  return color.alpha > 0
}